#include <stdlib.h>
#include <assert.h>
#include "cgstep.h"
#include "bool.h"
#include "myalloc.h"

#define EPSILON 1.e-12
static float* s;  /* model step */
static float* ss; /* residual step */
static bool alloc = FALSE; /* if s and ss are allocated */
static double dotprod (int n, const float* x, const float* y);

void cgstep( bool forget, int nx, int ny, float* x, const float* g, float* rr, 
  const float* gg) {
  double sds, gdg, gds, determ, gdr, sdr, alfa, beta;
  int i;
  if (alloc == FALSE) {
    alloc = forget = TRUE;
    s  = (float*) salloc (nx,sizeof (float));
    ss = (float*) salloc (ny,sizeof (float));
  }
  if ( forget == TRUE) {
    for (i = 0; i < nx; i++) s[i] = 0.;
    for (i = 0; i < ny; i++)  ss[i] = 0.;
    beta = 0.0;
    alfa = dotprod( ny, gg, gg);
    assert (alfa > 0.);
    alfa = - dotprod( ny, gg, rr) / alfa;
  } else {
    gdg = dotprod( ny, gg, gg);       
    sds = dotprod( ny, ss, ss);       
    gds = dotprod( ny, gg, ss);       
    if (gdg == 0. || sds == 0.) return;

    determ = 1.0 - (gds/gdg)*(gds/sds);
    if (determ > EPSILON) determ *= gdg * sds;
    else determ = gdg * sds * EPSILON;

    gdr = - dotprod( ny, gg, rr);
    sdr = - dotprod( ny, ss, rr);
    alfa = ( sds * gdr - gds * sdr ) / determ;
    beta = (-gds * gdr + gdg * sdr ) / determ;
  }
  for (i = 0; i < nx; i++) {
    s[i]  =  alfa * g[i] + beta *  s[i];
    x[i] +=  s[i];
  }
  for (i = 0; i < ny; i++) {
    ss[i] = alfa * gg[i] + beta * ss[i];
    rr[i] += ss[i];
  }
}

void cgstep_close (void) {
  if (alloc == TRUE) {
    free (s);
    free (ss);
    alloc = FALSE;
  }
}

static double dotprod (int n, const float* x, const float* y) {
  double prod;
  int i;
  prod = 0.;
  for (i = 0; i < n; i++) prod += x[i]*y[i];
  return prod;
}
