/*

dyamic.x pars <in.H trace_base=trace2.H  mat=mat.H path=path.H  >out.H  

Paramaters

stdin      - sepfile first trace for score
trace_base  - sepfile for the trace we are comparing to and keeping constant
stdout     - sepfile interpolated version of first trace
mat        - sepfile similarity and score matrix 
path       - sepfile Selected path through mat
nparts     - int     [20] Number of amp partitions for each trace 
sim        - int     [1]  similarity window in samples (1 or 2)
gap        - float   [0]  Gap penaly
match      - float   [1]  Match weight
mismatch   - float   [0]  Mismatch weight
ncoef      - int     [8]  Number of coeficeints for lagrange interpolation 
                          on each side (4 is sort of like sinc)
niter      - int     [n1/4] Number of iterations
eps        - float   [.2] Smoothing parameter

Description
Simple allignment through dynamic programing approach

*/

/*
renamed various to 'score matrix'
added mode flag to optionally output similarity matrix
tested that score matrix is unchanged on n1<->n2
added sim flag to do 2-pt similarity
removed mode flag added two array output for mat tag - Bob
*/
/*
Things to do, explore

Interpolation - Need to interpolate trace 2 to trace 1 coordinate system (or vis-vera)
                Should use sinc, or a higher order lagrange

Similarity  - should be able to be based on something more complex. This seems
              like the part that needs the most work

Score matrix - The gap penality is approriate for seismic data.  We should give
               a large penalty for large jumps, much smaller  for small jumps 
               Is this correct if we think in terms of faults ???? 

*/

/* Local functions*/
#include<sepmath.h>
#include<sepcube.h>
#include <solver.h>
#include<cgstep.h>
#include<iop.h>
#include<deriv.h>
#include <sinc_oper.h>
int grab_params(char*,char*,int*,int*,int*,float*,float*,float*,int*,int*,float*);
int read_convert(char*,int,int,float*,int*);
int build_score_mat(int,float,float,float,int,int*,int*,float**,float**);
int traceback_mat(int,float**,int**,int*);
int write_score_mat(char*,int,float**,float**);
int write_traceback(char*,int,int**);
int interp_trace(int,int,float,int**,int,int,float*,float*);
int write_trace(char*,int,float*);
float find_lagrange_interp(int n1,float x_out,float *x_in,float *y_in);


int main(int argc, char **argv, char **envp){
int n1; /*length of traces*/
int sim; /* flag to do 1 or 2 point similarity */
int nparts; /* number of parts to break it into*/
int   *cond1,*cond2; /*condensed form for the traces*/
float **sim_mat; /*similarity matrix*/
float **score_mat; /*score matrix*/
float match,mismatch,gap; /*match, mismatch and gap weights*/
int nlen; /*traverse length*/
int ncoef; /*number of coefs for lagrange */
int niter; /*(number of iterations*/
float eps; /*smoothing */
int **course ; /*course through score matrix*/
float *trace1,*trace2; /*the input traces*/
float *trace1_out,*trace2_out; /*the output traces*/

initpar(argc,argv); /*Initialize parameter handling*/
doc(SOURCE); /*Initialize self-documentation */

if(0!=grab_params("in","trace_base",&n1,&nparts,&sim,&match,&mismatch,&gap,&ncoef,&niter,&eps)) 
  seperr("trouble grabbing parameters ");

/*allocate space for the condensed form of the traces */
trace1=(float*) salloc(n1,sizeof(int));
trace2=(float*) salloc(n1,sizeof(int));
cond1=(int*) salloc(n1,sizeof(int));
cond2=(int*) salloc(n1,sizeof(int));

/*read in and convert the traces*/
if(0!=read_convert("in",n1,nparts,trace1,cond1)) 
  seperr("trouble reading in trace 1\n");
if(0!=read_convert("trace_base",n1,nparts,trace2,cond2)) 
  seperr("trouble reading in trace 2\n");

/*build 2-D similarity matrix*/
score_mat=(float**)salloc2(n1+1,n1+1,sizeof(float));
sim_mat=(float**)salloc2(n1+1,n1+1,sizeof(float));

if(0!=build_score_mat(n1,match,mismatch,gap,sim,cond1,cond2,sim_mat,score_mat))
  seperr("trouble building score matrix\n");

if(0!=write_score_mat("mat",n1,sim_mat,score_mat))
  seperr("trouble writting score matrix\n");

/*find best path*/
course=(int**) salloc2(2,n1*2,sizeof(int));
if(0!=traceback_mat(n1,score_mat,course,&nlen)) 
  seperr("trouble building traceback matrix\n");
if(0!=write_traceback("path",nlen,course)) 
  seperr("trouble writing traceback path\n");

/*interpolate the traces*/
trace1_out=(float*) salloc(n1,sizeof(int));
if(0!=interp_trace(nlen,n1,eps,course,ncoef,niter,trace1,trace1_out))
  seperr("trouble interpolating first trace \n");
/*if(0!=interp_trace(nlen,n1,course,1,ncoef,trace2,trace2_out))*/
/*  seperr("trouble interpolating second trace \n");*/
if(0!=write_trace("out",n1,trace1_out))
  seperr("trouble writing out trace 1\n");
/*if(0!=write_trace("trace2_out",nlen,trace2_out))*/
/*  seperr("trouble writing out trace 2\n");*/


return(0);
}

/* write out a trace */
int write_trace(char*tag,int n, float *trace){
  
auxputch("n1","d",&n,tag) ; /*set the new trace length*/
if(n*4!=srite(tag,trace,n*4)) return(1); /*write out trace*/
return(0);
}


/*interpolate the traces -
   for now we are just going to stretch and squueze both
   a better solution would be to interpolate trace two
   to the coordinates implied by a trace1 
   This scheme will make the o and d parameters meaningless */
/*int interp_trace(int,int,float,int**,int,int,float*,float*);*/
/*if(0!=interp_trace(nlen,n1,eps,course,ncoef,niter,trace1,trace1_out))*/

int interp_trace(int npath, int n1,float eps,int **course, int ncoef,int niter, float *in,float *out){
float  *t_hold;
int *i_have;
float f,b,e;
int ib,ie,i,ilast_e,ilast_b,j,ibeg,iend,nc;
float *base_equiv,*first_pos,*x_in,*y_in;

/*array saying what input locations corresponds to a given output location*/

base_equiv=(float*) salloc(n1,sizeof(int));
/*slow approach, stupid approach, but who cares, we
   are dealing with small problems*/


/*loop over the base trace locations*/
for(i=0; i <n1 ; i++){
  /* check for a boundary condition */
  if(course[npath-1][1]-1 >i ){
     /*our trace starts before the base so zero it*/
     base_equiv[i]=-1.; /*notify interp to put a zero*/
  }
  else if(course[0][1]-1 < i){
     base_equiv[i]=-1.;
  }
  else{
     /*find the first and last instance of this location in the course func*/
     ie=0; while(ie < npath && course[ie][1] != i) ie++;
     ib=npath; while(ib >0 && course[ib][1]  != i) ib--;
     if(ib==ie) base_equiv[i]=course[ib][0];
     else{
       base_equiv[i]=(course[ib][0]+course[ie][0])*.5 ;
     }
  }
} 
/*the corrorally to base_equiv (think push rather than pull*/
/*array saying what output location corresponds to a given input location*/
first_pos=(float*) salloc(n1,sizeof(int));
ibeg=0; iend=n1-1;
if(base_equiv[0]>0) {
  for(i=0; i < base_equiv[0];i++) first_pos[i]=-1;
  ibeg=base_equiv[0];
}
if(base_equiv[n1-1]>0 && base_equiv[n1-1] <  n1-1) {
  for(i=base_equiv[0]+1;i<n1;i++) first_pos[i]=-1;
  iend=base_equiv[n1-1];
}
for(i=ibeg; i <=iend;i++){ 
     ib=0; while(ib < n1 && base_equiv[ib]<i ) ib++;
     first_pos[i]=((float)(i-base_equiv[ib-1]))/((float)(base_equiv[ib]-base_equiv[ib-1]))+ib-1;
}

/* time to  do */

 sinc_oper_init(ncoef,n1,first_pos);
 solver_reg(sinc_op,cgstep,i_op,n1,n1,out,n1,in,niter,eps);
 sinc_clean();

/*
x_in=(float*) salloc(ncoef*2,sizeof(int));
y_in=(float*) salloc(ncoef*2,sizeof(int));
 for(i=0; i < n1;i++){
   if(base_equiv[i]<0.) out[i]=0.;
   else{
     ibeg=base_equiv[i];
     nc=0;
     for(ib=ibeg,j=0; j< ncoef ;j++,ib--){
       if(ib<0){
         if(j==0) seperr("should be unreachable1\n");
         x_in[j]=x_in[j-1]-.001; y_in[j]=y_in[j-1];
       }
       else{
         if(first_pos[ib]<0.){
            if(j==0){
              x_in[j]=x_in[0]-.001; y_in[j]=y_in[0];
            }
            x_in[j]=x_in[j-1]-.001; y_in[j]=y_in[j-1];
         }
         else {
          x_in[nc]=first_pos[ib];y_in[nc]=in[ib];nc++;
         }
       }
     }
     for(ib=ibeg+1,j=ncoef; j< 2*ncoef ;j++,ib++){
       if(ib>n1-1){
         if(j==ncoef){ x_in[j]=x_in[0]; y_in[j]=y_in[0]; }
         x_in[j]=x_in[j-1]-.001; y_in[j]=y_in[j-1];
       }
       else{
         if(first_pos[ib]<0.){
            if(j==ncoef) {x_in[j]=x_in[0]; y_in[j]=y_in[0];}
            x_in[j]=x_in[j-1]-.001; y_in[j]=y_in[j-1];
         }
         else {x_in[nc]=first_pos[ib];y_in[nc]=in[ib];nc++;}
       }
     }
     out[i]=find_lagrange_interp(nc,(float)i,x_in,y_in);
     ib=base_equiv[i]; f=base_equiv[i]-ib;
     if(ib >= n1) { ib=n1-2; f=1.;}
     out[i]=in[ib]*(1.-f)+f*in[ib+1]+1;
   }
 }
free(x_in);free(y_in);
  */
 free(base_equiv);
free(first_pos);
 return(0);
}

float find_lagrange_interp(int n1,float x_out,float *x_in, float *y_in){
int i,j;
float wt,a;

 a=0;
 for(i=0; i < n1; i++){
   wt=1.;
   for(j=0;  j < n1; j++){ 
     if(i!=j ) wt=wt*(x_out-x_in[j])/(x_in[i]-x_in[j]);
   }
   a=a+wt*y_in[i];
 } 
 return(a);
}

int write_traceback(char *tag, int n1, int **course){
float pt[2];
int i;
auxputch("n1","d",&n1,tag);
i=8; auxputch("esize","d",&i,tag);

for(i=0; i < n1; i++){
  pt[0]=course[i][0]; pt[1]=course[i][1];
  if(8!=srite(tag,pt,8)) {
     fprintf(stderr,"wrote the wrong number bytes for %s \n",tag);
     return(1);
  }
}
  return(0);
}


/*traceback matrix*/
int traceback_mat(int n, float **mat,int **course,int *nlen){
int i,iloc,imax,npath,i1,i2,ix,iy;
float diag,l1,l2;

/*[i][0] represents location in trace 1*/
/*[i][1] represents the location in trace 2*/
/*find our starting position*/

/*find the maximum value along the bottom row and collumn */
npath=0;
course[0][1]=n; course[0][0]=n;
for(i=n-1;i>0;i--){
  if(imax<mat[i][n]){ 
    imax=mat[i][n]; course[npath][0]=i; course[npath][1]=n; }
}
for(i=n-1;i>0;i--){
  if(imax<mat[n][i]){ imax=mat[n][i]; course[npath][0]=n; course[npath][1]=i; }
}

         /*now its time  to traceback*/
while(course[npath][0] >0 && course[npath][1] >0){ /*loop until we reach edge*/
  i1=course[npath][0]; i2=course[npath][1];
  diag=mat[i1-1][i2-1];
  l1  =mat[i1-1][i2];
  l2  =mat[i1][i2-1];
  /*if diagonal is equal or better than others choose it*/
  if(diag >= l1 && diag >= l2){ 
    npath++; course[npath][0]=i1-1; course[npath][1]=i2-1;
  }
  /*we prefer to skip along the second trace*/
  else if(l2>= diag && l2 >= l1){
    npath++; course[npath][0]=i1;course[npath][1]=i2-1;
  }
  else{
    npath++; course[npath][0]=i1-1;course[npath][1]=i2;
  }
}

*nlen=npath;
return(0);
}


/*write out a  score matrix*/
int write_score_mat(char *tag, int n1, float **sim, float **score){
int i1,i2;
float d;

 /*write out axis information*/
 i1=1;d=1.;sep_put_data_axis_par(tag,&i1,&n1,&d,&d,"trace 1");
 i1=2;d=1.;sep_put_data_axis_par(tag,&i1,&n1,&d,&d,"trace 2");
 i2=2;i1=3;d=1.;sep_put_data_axis_par(tag,&i1,&i2,&d,&d,"matrix");
 

  /*write out the matrix*/
  for(i1=1;i1<=n1;i1++){ 
     if(4*n1!=srite(tag,&(sim[i1][1]),4*n1)){
       fprintf(stderr,"trouble writing matrix row %d  to tag %s \n",i1,
         tag); return(1);
     } 
  }
  for(i1=1;i1<=n1;i1++){ 
     if(4*n1!=srite(tag,&(score[i1][1]),4*n1)){
       fprintf(stderr,"trouble writing matrix row %d  to tag %s \n",i1,
         tag); return(1);
     } 
  }


return(0);

}
/*
        if(cond1[i1]-cond1[i1-1]==cond2[i2]-cond2[i2-1]) val=match;
*/
/* Build score or similarity matrix */
int build_score_mat(int n,float match,float mis,float gap,int sim,int *cond1,int *cond2,float **sim_mat,float **score_mat){
int i2,i1;
float val;
for(i1=0;i1<=n;i1++){ sim_mat[i1][0]=sim_mat[0][i1]=0;}
for(i1=0;i1<=n;i1++){ score_mat[i1][0]=score_mat[0][i1]=0;}
/* verified results are symmetric 1->n or n->1 */
for(i2=1;i2<=n;i2++){
   for(i1=1;i1<=n;i1++){
      if(sim==1) { 
        if(cond1[i1]==cond2[i2]) val=match;
        else val=mis;
      }
      if(sim==2) { 
        if(cond1[i1]==cond2[i2] && cond1[i1-1]==cond2[i2-1]) val=match;
        else val=mis;
      }
      score_mat[i1][i2]=MAX(MAX(score_mat[i1-1][i2-1]+val,
      score_mat[i1-1][i2]+gap), score_mat[i1][i2-1]+gap);
      sim_mat[i1][i2]=val;
   }
}

return(0);
}

/* read in trace and partition amp into nparts */
int read_convert(char *tag, int n1,int nparts, float*trace, int *cond){
int i1;
float  min,max,interval;


if(n1*4!=sreed(tag,trace,n1*4)) {
 fprintf(stderr,"trouble reading %d bytes from tag %s \n",n1*4,tag);
 return(1); 
}

/*find the minimum and maximum in the trace */
min=max=trace[0];
for(i1=1;i1< n1;i1++){
  if(min > trace[i1]) min=trace[i1];
  if(max < trace[i1]) max=trace[i1];
}
interval=(max-min)/(nparts-1);

/* reduce the dyamic range nparts */
for(i1=0;i1< n1;i1++)
  cond[i1]=(int)((trace[i1]-min)/interval);
return(0);
}

/*grab parameters*/
int grab_params(char *tag1, char *tag2, int *n1, int *nparts,  int *sim, float *match, float *mis,float *gap,int *ncoef, int *niter, float *eps){
int n,esize;

/*look for n1 (length of trace) in the history files specified
   by the given tags, if not found give an error*/
   
if(1!=auxpar("n1","d",n1,tag1) || 1!=auxpar("n1","d",&n,tag2)){
  fprintf(stderr,"could not find n1 in both tag(%s) and tag(%s) \n",
    tag1,tag2);
  return(1);
}

/*make sure the length of traces are the same*/
if(*n1!=n){
  fprintf(stderr,"length of traces not the same");
  return(1);
}

/*make sure the input is real data in 4 byte representation*/

/*look for esize if not found, set it to 4*/
if(1!=auxpar("esize","d",&n,tag2)) esize=4;
if(1!=auxpar("esize","d",&esize,tag1)) esize=4;
if(n!=4 || esize!=4){
 fprintf(stderr,"esize is not 4 in %s or %s \n",tag1,tag2);
 return(1);
}


/*grab nparts and weights from the command line if not found set to 20*/
if(1!=getch("nparts","d",nparts)) *nparts=20;
if(1!=getch("sim","d",sim))       *sim=1;
if(1!=getch("match","f",match))   *match=1;
if(1!=getch("mismatch","f",mis))  *mis=0;
if(1!=getch("ncoef","d",ncoef))       *ncoef=8;
if(1!=getch("niter","d",niter))       *niter=*n1;
if(1!=getch("gap","f",gap))       *gap=0;
if(1!=getch("eps","f",eps))       *eps=.4;

putch("ncoef","d",ncoef); /*write to stdout history file nparts*/
putch("niter","d",niter); /*write to stdout history file nparts*/
putch("nparts","d",nparts); /*write to stdout history file nparts*/
putch("match","f",match); /*write to stdout history file match*/
putch("mismatch","f",mis); /*write to stdout history file mismatch*/
putch("gap","f",gap); /*write to stdout history file gap*/
putch("eps","f",eps); /*write to stdout history file gap*/

return(0);

}
